import numpy as np
import pandas as pd
from sklearn.preprocessing import MinMaxScaler, StandardScaler


def process_data(df):
    """
    Process data for ML model.
    :param df: pandas.DataFrame - raw dataset to analyze
    :return: pandas.DataFrame with processed data
    """
    # parse data
    func = lambda x: 1 if len(x) > 0 else np.NAN
    try_df = df.copy()
    try_df['Data'] = try_df['Data'].apply(eval)
    try_df['Data_2'] = try_df['Data_2'].apply(eval)

    try_df['check'] = try_df['Data'].apply(func)
    try_df = try_df.dropna()
    try_df['check'] = try_df['Data_2'].apply(func)
    try_df = try_df.dropna().drop(['check'], axis=1).reset_index(drop=True)

    photo = try_df.drop(['Data_2'], axis=1)
    piezo = try_df.drop(['Data'], axis=1)

    scaler_photo = MinMaxScaler()
    flag_photo = True
    for i, row in photo.iterrows():
        try:
            if flag_photo:
                photo.loc[i, 'Data'] = scaler_photo.fit_transform(np.array(row['Data']).reshape(-1, 1))
                flag_photo = False
            else:
                photo.loc[i, 'Data'] = scaler_photo.transform(np.array(row['Data']).reshape(-1, 1))
        except:
            continue
    # normalize data
    scaler_piezo = MinMaxScaler()
    flag_piezo = True
    for i, row in piezo.iterrows():
        try:
            if flag_piezo:
                piezo.loc[i, 'Data_2'] = scaler_piezo.fit_transform(np.array(row['Data_2']).reshape(-1, 1))
                flag_piezo = False
            else:
                piezo.loc[i, 'Data_2'] = scaler_piezo.transform(np.array(row['Data_2']).reshape(-1, 1))
        except:
            continue

    all_df = photo
    all_df['Data_2'] = piezo['Data_2']
    percent_df = all_df.copy()

    # quantization of data
    quant_list = [0.01, 0.2, 0.25, 0.3, 0.4, 0.5, 0.6, 0.7, 0.75, 0.8, 0.9, 0.99]

    def _quant(row: pd.Series, quant=None):
        quant = np.quantile(row, quant)
        return quant

    for q in quant_list:
        try:
            percent_df[f'Data_q_{q}'] = percent_df['Data'].apply(_quant, quant=q)
            percent_df[f'Data2_q_{q}'] = percent_df['Data_2'].apply(_quant, quant=q)
        except:
            continue

    stats_df = percent_df.copy()
    for f, v in zip([np.mean, np.max, np.min, np.median], ["mean", "max", "min", "median"]):
        try:
            stats_df[f'Data_{v}'] = stats_df['Data'].apply(f)
            stats_df[f'Data2_{v}'] = stats_df['Data_2'].apply(f)
        except:
            continue

    prep_df = stats_df.astype({
        'Test_index': 'int',
        'Presentation': 'int',
        'Question': 'int'
    })

    # spectrum analyze of data

    N = 10

    def abs_amplitude(row: pd.Series):
        """
        Calculate absolute amplitude of signal
        :param row: pandas.Series - signal
        :return: pandas.Series - absolute amplitude
        """
        F = np.fft.rfft(row, n=None, axis=-1)
        A = [((F[i].real) ** 2 + (F[i].imag) ** 2) ** 0.5 for i in np.arange(0, N, 1)]
        return A

    def phase(row: pd.Series):
        """
        Calculate phase of signal. Fourier transform of signal is used.
        :param row: pandas.Series - signal
        :return: pandas.Series - phase
        """
        F = np.fft.rfft(row, n=None, axis=-1)
        arg = []
        for i in np.arange(0, N, 1):
            if F[i].imag != 0:
                t = (-np.tanh((F[i].real) / (F[i].imag)))
                arg.append(t)
            else:
                arg.append(np.pi / 2)
        return arg

    # generate new features
    prep_df['data_apm'] = prep_df['Data'].apply(abs_amplitude)
    prep_df['data_2_apm'] = prep_df['Data_2'].apply(abs_amplitude)
    prep_df['data_phase'] = prep_df['Data'].apply(phase)
    prep_df['data_2_phase'] = prep_df['Data_2'].apply(phase)
    sensor_1 = prep_df['data_2_apm'].apply(pd.Series)
    sensor_1.columns = [f'data_2_amp_{i}' for i in range(sensor_1.shape[1])]
    sensor_2 = prep_df['data_2_phase'].apply(pd.Series)
    sensor_2.columns = [f'data_2_phase_{i}' for i in range(sensor_2.shape[1])]

    sensor_3 = prep_df['data_apm'].apply(pd.Series)
    sensor_3.columns = [f'data_amp_{i}' for i in range(sensor_3.shape[1])]
    sensor_4 = prep_df['data_phase'].apply(pd.Series)
    sensor_4.columns = [f'data_phase_{i}' for i in range(sensor_4.shape[1])]

    sensor_df = pd.concat([prep_df, sensor_1, sensor_2, sensor_3, sensor_4], axis=1)
    try_set = sensor_df.drop(['Data_2', 'Data', 'Filename',
                              'data_2_apm', 'data_2_phase', 'data_apm', 'data_phase'], axis=1)
    return try_set


def process_data_nn(df):
    """
    Process data for neural network
    :param df: Pandas dataframe - dataframe with raw data
    :return: numpy array with processed data
    """
    correct_mask = (df.Data.apply(lambda x: len(eval(x))) == 240) & (df.Data_2.apply(lambda x: len(eval(x))) == 240)

    vals = df.loc[correct_mask].Data.apply(eval)
    res = [val for val in vals if len(val) == 240]
    data = np.array(res)
    sc = StandardScaler()
    transformed_data_1 = sc.fit_transform(data)

    vals = df.loc[correct_mask].Data_2.apply(eval)
    res = [val for val in vals if len(val) == 240]
    data = np.array(res)
    sc = StandardScaler()
    transformed_data_2 = sc.fit_transform(data)

    transformed_data = np.concatenate((transformed_data_1, transformed_data_2), axis=1)
    return transformed_data



import pickle
import numpy as np
import torch
import torch.nn as nn
import overrated_ld_stress_detector.preprocessing as preprocessing
from scipy import stats
from catboost import CatBoostClassifier
from sklearn.model_selection import StratifiedKFold


class SignalModel(nn.Module):
    """
    Define architecture of the used model
    """
    def __init__(self):
        """
        Initialize the model with Convolution 1d layers
        """
        super().__init__()
        self.conv_net1 = nn.Sequential(
            nn.Conv1d(1, 64, 40, padding='same'),
            nn.LeakyReLU(),
            nn.Conv1d(64, 128, 40, padding='same'),
            nn.LeakyReLU(),
            nn.Conv1d(128, 64, 40, padding='same'),
            nn.LeakyReLU(),
            nn.MaxPool1d(240),
            nn.Flatten(),
        )

        self.conv_net2 = nn.Sequential(
            nn.Conv1d(1, 64, 40, padding='same'),
            nn.LeakyReLU(),
            nn.Conv1d(64, 128, 40, padding='same'),
            nn.LeakyReLU(),
            nn.Conv1d(128, 64, 40, padding='same'),
            nn.LeakyReLU(),
            nn.MaxPool1d(240),
            nn.Flatten()
        )

        self.proc_net = nn.Sequential(
            nn.Linear(128, 64),
            nn.LeakyReLU(),
            nn.Linear(64, 3)
        )

    def forward(self, x):
        """
        Forward pass of the model
        """
        x_1 = x[..., 0:240]
        x_2 = x[..., 240:]
        conv_out1 = self.conv_net1(x_1)
        conv_out2 = self.conv_net2(x_2)

        out_c = torch.concat((conv_out1, conv_out2), dim=-1)
        return self.proc_net(out_c)


class PytorchModel:
    """
    Convoluted neural network model for predicting the stress level of a signal.
    """
    def __init__(self,
                 model_path='models/nn_full.pth',
                 device: str = "cpu") -> None:
        """
        :param model_path: str, path to model file
        :param device: str, device to use ['cpu', 'cuda']
        """
        if model_path is None:
            model_path = 'models/nn_full.pth'
        self.model = SignalModel()
        self.model.load_state_dict(torch.load(model_path, map_location=torch.device(device)))
        self.device = device

    def predict(self, data):
        """
        Predict results by blending predictions from all models.
        :param data: pandas.DataFrame - raw dataset to analyze
        :return: pandas.DataFrame with predicted results
        """
        self.model.eval()
        prepared_data = torch.tensor(preprocessing.process_data_nn(data), dtype=torch.float32).to(self.device).unsqueeze(1)
        return self.model(prepared_data).cpu().detach().argmax(1).numpy().flatten()


class CatboostModel:
    """
    Gradient boosting catboost model
    """
    def __init__(self,
                 model_path='models',
                 model_count=5) -> None:
        """
        Load and initialize models from files.
        :param model_path: str, path to folder with models. Models name should be 'model_<number>.pkl'
        :param model_count: int, number of models to load
        """
        if model_path is None:
            model_path = 'models'
        self.models = [pickle.load(open(model_path + '/model_' + str(i) + ".pckl", 'rb')) for i in range(model_count)]
        self.model_count = model_count

    def predict(self, df):
        """
        Predict results by blending predictions from all models.
        :param df: pandas.DataFrame - raw dataset to analyze
        :return: pandas.DataFrame with predicted results
        """
        modified_data = preprocessing.process_data(df)
        result = np.array([self.models[i].predict(modified_data).flatten() for i in range(0, self.model_count)])
        return stats.mode(result)[0][0]

    def train(self,
              df,
              model_count=5,
              save_path='models/',
              iterations=1000,
              ):
        """
        Train model on data.
        :param df: pandas.DataFrame - raw dataset to train on
        :param model_count: int, number of models to train
        :param save_path: str, path to folder to save models
        :param iterations: int, number of iterations to train
        """
        modified_data = preprocessing.process_data(df)
        target = modified_data['Class_label']
        modified_data = modified_data.drop(['Class_label'], axis=1)
        cat_features = modified_data.select_dtypes(include=['category']).columns.to_list()

        skfg = StratifiedKFold(n_splits=model_count, random_state=142, shuffle=True)
        models = []
        for train_ids, val_ids in skfg.split(modified_data, target):
            classifier = CatBoostClassifier(random_state=142, iterations=iterations,
                                            verbose=250, use_best_model=True,
                                            eval_metric='TotalF1')
            classifier.fit(modified_data.loc[train_ids], target[train_ids], cat_features=cat_features,
                           eval_set=(modified_data.loc[val_ids], target[val_ids]))
            models.append(classifier)

        if save_path is not None:
            for i, model in enumerate(models):
                pickle.dump(model, open(save_path + 'model_{i}.pckl', 'wb'))
        return models



import plotly.express as px
import numpy as np


def visualize_data(df,
                   user_id,
                   test_id,
                   presentation_id,
                   result=None,
                   frequency=20,
                   len_seconds=12):
    """
    Visualize data.
    df - pandas dataframe
    user_id - user id
    test_id - test id
    presentation_id - presentation id
    """

    vals = df[df.Filename == f'{user_id}'].query(f'Test_index=={test_id}').query(
        f'Presentation=={presentation_id}').Data.apply(eval)
    res = [val for val in vals if len(val) == frequency * len_seconds]
    data = np.array(res)

    vals2 = df[df.Filename == f'{user_id}'].query(f'Test_index=={test_id}').query(
        f'Presentation=={presentation_id}').Data_2.apply(eval)
    res2 = [val for val in vals2 if len(val) == frequency * len_seconds]
    data2 = np.array(res2)

    if result is None:
        target = df[df.Filename == f'{user_id}'].query(f'Test_index=={test_id}').query(
            f'Presentation=={presentation_id}').Class_label.values
    else:
        target = result

    col_dict = {0: 'green', 1: 'yellow', 2: 'red'}

    fig = px.line(y=data.flatten())
    fig.update_layout(title=f'Фотоплетизмограмма для всех вопросов пользователю {user_id} '
                            f'для группы вопросов {test_id} с {presentation_id} номером повторения.\n',
                      legend_title='Классы вопросов',
                      xaxis_title='Время',
                      yaxis_title='Амплитуда сигнала',
                      )
    fig.update_layout()

    for i in range((frequency * len_seconds), len(data.flatten()) + 1, frequency * len_seconds):
        fig.add_vrect(x0=i - frequency * len_seconds,
                      x1=i,
                      fillcolor=col_dict[target[i // (frequency * len_seconds) - 1]],
                      opacity=0.2,
                      annotation={"text": str(int(target[i // (frequency * len_seconds) - 1])),
                                  "showarrow": False},
                      )

    fig.show()

    fig = px.line(y=data2.flatten())
    fig.update_layout(title=f'Пьезоплетизмограмма для всех вопросов пользователю {user_id}'
                            f'для группы вопросов {test_id} с {presentation_id} номером повторения',
                      xaxis_title='Время',
                      yaxis_title='Амплитуда сигнала')
    for i in range((frequency * len_seconds), len(data2.flatten()) + 1, frequency * len_seconds):
        fig.add_vrect(x0=i - frequency * len_seconds,
                      x1=i,
                      fillcolor=col_dict[target[i // (frequency * len_seconds) - 1]],
                      opacity=0.2,
                      annotation={"text": str(int(target[i // (frequency * len_seconds) - 1])),
                                  "showarrow": False})
    fig.show()



import pandas as pd

from ml import pretrained_model
from visualization import utils

df = pd.read_excel(r"models/dataset_train.xlsx", engine='openpyxl')

utils.visualize_data(df,
                     user_id='8fc79c7f-bbdb-4512-b460-c75aacd1a3c7',
                     test_id=4,
                     presentation_id=1)

df = df.drop("Class_label",
             axis=1)

model = pretrained_model.get_model(model_type='catboost')

result = model.predict(df)

utils.visualize_data(df,
                     user_id='8fc79c7f-bbdb-4512-b460-c75aacd1a3c7',
                     test_id=3,
                     presentation_id=1,
                     result=result)



import os
import re
import sys
from pathlib import Path
from shutil import rmtree
from typing import List, Tuple

import setuptools
from setuptools import Command, find_packages, setup

# Package meta-data.
name = "overrated_ld_stress_detector"
description = "Stress detection."
url = "https://github.com/mansasha21/overrated-ld-stress-detector"
email = "no@gmail.com"
author = "Overrated team"
requires_python = ">=3.0.0"
current_dir = Path(__file__).absolute().parent


def get_version() -> str:
    version_file = current_dir / "overrated_ld_stress_detector" / "__init__.py"
    with version_file.open(encoding="utf-8") as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)  # type: ignore


# What packages are required for this module to be executed?
try:
    with (current_dir / "requirements.txt").open(encoding="utf-8") as f:
        required = f.read().split("\n")
except FileNotFoundError:
    required = []

# What packages are optional?
extras = {"test": ["pytest"]}

version = get_version()

about = {"__version__": version}


def get_test_requirements() -> List[str]:
    requirements = ["pytest"]
    if sys.version_info < (3, 3):
        requirements.append("mock")
    return requirements


def get_long_description() -> str:
    return ""


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options: List[Tuple] = []

    @staticmethod
    def status(s: str) -> None:
        """Print things in bold."""
        print(s)  # noqa: T001

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(current_dir, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system(f"{sys.executable} setup.py sdist bdist_wheel --universal")

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        self.status("Pushing git tags...")
        os.system(f"git tag v{about['__version__']}")
        os.system("git push --tags")

        sys.exit()


setup(
    name=name,
    version=version,
    description=description,
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Overrated team",
    license="MIT",
    url=url,
    include_package_data=True,
    packages=setuptools.find_packages(),
    data_files=[("overrated_ld_stress_detector",
                 ["overrated_ld_stress_detector/models/nn_full.pth",
                  "overrated_ld_stress_detector/models/model_4.pckl",
                  "overrated_ld_stress_detector/models/model_3.pckl",
                  "overrated_ld_stress_detector/models/model_2.pckl",
                  "overrated_ld_stress_detector/models/model_1.pckl",
                  "overrated_ld_stress_detector/models/model_0.pckl",
                  "overrated_ld_stress_detector/models/dataset_test.xlsx",
                  "overrated_ld_stress_detector/README.md",
                  "overrated_ld_stress_detector/models/dataset_train.xlsx"])],
    install_requires=required,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    cmdclass={"upload": UploadCommand},
)


from overrated_ld_stress_detector.ml.predict import PytorchModel
from overrated_ld_stress_detector.ml.predict import CatboostModel


def get_model(model_type='catboost',
              model_path=None,
              device='cpu'):
    """
    Returns a model object: convolutional neural network or gradient boosting catboost.
    :param model_type: str, one of ['cnn', 'catboost']
    :param model_path: str, path to model file
    :param device: str, device to use ['cpu', 'cuda']
    """
    model = None
    if model_type == 'catboost':
        model = CatboostModel(model_path)
    elif model_type == 'cnn':
        model = PytorchModel(device=device)
    return model
